/*
 * DirectDraw XVidMode interface
 *
 * Copyright 2001 TransGaming Technologies, Inc.
 */

#include "config.h"
#include <string.h>

#include "ts_xlib.h"
#include "ts_xf86vmode.h"
#include "x11drv.h"
#include "x11ddraw.h"
#include "xvidmode.h"

#include "windef.h"
#include "wingdi.h"
#include "ddrawi.h"
#include "debugtools.h"

#include "sysmetrics.h"

DEFAULT_DEBUG_CHANNEL(x11drv);

#ifdef HAVE_LIBXXF86VM

extern int usexvidmode;

extern void X11DRV_EnableGrabs(void);
extern void X11DRV_DisableGrabs(void);
static int xf86vm_in_abnormalmode = 0;

int app_resolution_change;

static int xf86vm_event, xf86vm_error, xf86vm_major, xf86vm_minor;

#ifdef X_XF86VidModeSetGammaRamp
static int xf86vm_gammaramp_size;
static BOOL xf86vm_use_gammaramp;
#endif

LPDDHALMODEINFO xf86vm_modes;
unsigned xf86vm_mode_count;
XF86VidModeModeInfo** modes;

static void convert_modeinfo( const XF86VidModeModeInfo *mode, LPDDHALMODEINFO info )
{
  info->dwWidth      = mode->hdisplay;
  info->dwHeight     = mode->vdisplay;
  if (mode->htotal!=0 && mode->vtotal!=0)
      info->wRefreshRate = mode->dotclock * 1000 / (mode->htotal * mode->vtotal);
  else
      info->wRefreshRate = 0;
  TRACE(" width=%ld, height=%ld, refresh=%d\n",
        info->dwWidth, info->dwHeight, info->wRefreshRate);
  /* XVidMode cannot change display depths... */
  /* let's not bother with filling out these then... */
  info->lPitch         = 0;
  info->dwBPP          = 0;
  info->wFlags         = 0;
  info->dwRBitMask     = 0;
  info->dwGBitMask     = 0;
  info->dwBBitMask     = 0;
  info->dwAlphaBitMask = 0;
}

static void convert_modeline(int dotclock, const XF86VidModeModeLine *mode, LPDDHALMODEINFO info)
{
  info->dwWidth      = mode->hdisplay;
  info->dwHeight     = mode->vdisplay;
  if (mode->htotal!=0 && mode->vtotal!=0)
      info->wRefreshRate = dotclock * 1000 / (mode->htotal * mode->vtotal);
  else
      info->wRefreshRate = 0;
  TRACE(" width=%ld, height=%ld, refresh=%d\n",
        info->dwWidth, info->dwHeight, info->wRefreshRate);
  /* XVidMode cannot change display depths... */
  /* let's not bother with filling out these then... */
  info->lPitch         = 0;
  info->dwBPP          = 0;
  info->wFlags         = 0;
  info->dwRBitMask     = 0;
  info->dwGBitMask     = 0;
  info->dwBBitMask     = 0;
  info->dwAlphaBitMask = 0;
}

void X11DRV_XF86VM_Init(void)
{
  int nmodes, i;

  if (xf86vm_major) return; /* already initialized? */

  if (!usexvidmode) return;

  /* see if XVidMode is available */
  if (!TSXF86VidModeQueryExtension(gdi_display, &xf86vm_event, &xf86vm_error)) return;
  if (!TSXF86VidModeQueryVersion(gdi_display, &xf86vm_major, &xf86vm_minor)) return;

#ifdef X_XF86VidModeSetGammaRamp
  if (xf86vm_major > 2 || (xf86vm_major == 2 && xf86vm_minor >= 1))
  {
      TSXF86VidModeGetGammaRampSize(gdi_display, DefaultScreen(gdi_display),
				  &xf86vm_gammaramp_size);

      if (xf86vm_gammaramp_size == 256)
	  xf86vm_use_gammaramp = TRUE;
  }
#endif

  /* if in desktop mode, don't do mode switching */
  if (root_window != DefaultRootWindow(gdi_display)) return;

  /* retrieve modes */
  if (!TSXF86VidModeGetAllModeLines(gdi_display, DefaultScreen(gdi_display), &nmodes,
				    &modes))
    return;

  TRACE("XVidMode modes: count=%d\n", nmodes);

  xf86vm_mode_count = nmodes;
  xf86vm_modes = HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, sizeof(DDHALMODEINFO) * nmodes);

  /* convert modes to DDHALMODEINFO format */
  for (i=0; i<nmodes; i++)
    convert_modeinfo(modes[i], &xf86vm_modes[i]);

  TRACE("Enabling XVidMode\n");
}

void X11DRV_XF86VM_Cleanup(void)
{
  if (modes) TSXFree(modes);
}

int X11DRV_XF86VM_GetCurrentMode(void)
{
  XF86VidModeModeLine line;
  int dotclock, i;
  DDHALMODEINFO cmode;

  if (!xf86vm_modes) return 0; /* no XVidMode */

  TRACE("Querying XVidMode current mode\n");
  TSXF86VidModeGetModeLine(gdi_display, DefaultScreen(gdi_display), &dotclock, &line);
  convert_modeline(dotclock, &line, &cmode);
  for (i=0; i<xf86vm_mode_count; i++)
    if (memcmp(&xf86vm_modes[i], &cmode, sizeof(cmode)) == 0) {
      TRACE("mode=%d\n", i);
      return i;
    }
  ERR("unknown mode, shouldn't happen\n");
  return 0; /* return first mode */
}

void X11DRV_XF86VM_SetCurrentMode(int mode)
{
  RECT newDesktopRect;
  HWND desktopWindow;
  if (!xf86vm_modes) return; /* no XVidMode */

  TRACE("xvidmode, switching to mode %i\n", mode);
  TSXF86VidModeSwitchToMode(gdi_display, DefaultScreen(gdi_display), modes[mode]);
#if 0 /* it is said that SetViewPort causes problems with some X servers */
  TSXF86VidModeSetViewPort(gdi_display, DefaultScreen(gdi_display), 0, 0);
#else
  TSXWarpPointer(gdi_display, None, DefaultRootWindow(gdi_display), 0, 0, 0, 0, 0, 0);
#endif
  TSXSync(gdi_display, False);

  screen_width = xf86vm_modes[mode].dwWidth;
  screen_height = xf86vm_modes[mode].dwHeight;
  
  /* change the desktop window to reflect the new screen size.
   * Note: This doesnt fix problems with mouse warping to centre of screen
   * rather than desktop if XVidMode is turned off..
   */
  desktopWindow = GetDesktopWindow();
  SetRect( &newDesktopRect, 0, 0, screen_width, screen_height );
  WIN_SetRectangles( desktopWindow, &newDesktopRect, &newDesktopRect );
}

BOOL X11DRV_EnumDisplayModes(LPDEVMODEA devmode, DWORD n)
{
  int vidmode_count, num_depths;
  XF86VidModeModeInfo** vidmodes;
  int current_depth;

  TRACE("requested mode %ld\n", n);

  if (n == -1 && root_window != DefaultRootWindow(gdi_display))
  { /* if requested current mode and in desktop mode - special case */
    devmode->dmDisplayFlags = 0;
    /* FIXME: devmode->dmDisplayFrequency */
    devmode->dmBitsPerPel = GetSystemMetrics(SM_WINE_BPP);
    devmode->dmPelsHeight = GetSystemMetrics(SM_CYSCREEN);
    devmode->dmPelsWidth = GetSystemMetrics(SM_CXSCREEN);
    devmode->dmFields = DM_BITSPERPEL | DM_PELSWIDTH |
                        DM_PELSHEIGHT | DM_DISPLAYFLAGS;
    return TRUE;
  }
  if (n == -1) n = xf86vm_in_abnormalmode;
  if (!TSXF86VidModeGetAllModeLines(gdi_display, DefaultScreen(gdi_display), &vidmode_count,
                                    &vidmodes))
    return FALSE;
  if (GetSystemMetrics(SM_WINE_BPP) == 24)
    num_depths = 2;
  else num_depths = 1;
  TRACE("\tof %i\n", vidmode_count);
  if (n >= (vidmode_count * num_depths))
  {
    TSXFree(vidmodes);
    return FALSE;
  }

  if (n >= vidmode_count)
  {
    current_depth = 16;
    n = n % vidmode_count;
  }
  else
    current_depth = GetSystemMetrics(SM_WINE_BPP);
  /* fill the requested data */
  devmode->dmDisplayFlags = 0;
  devmode->dmDisplayFrequency =
          vidmodes[n]->dotclock * 1000 / (vidmodes[n]->htotal * vidmodes[n]->vtotal);
  devmode->dmBitsPerPel = current_depth;
  devmode->dmPelsHeight = vidmodes[n]->vdisplay;
  devmode->dmPelsWidth = vidmodes[n]->hdisplay;
  devmode->dmFields = DM_BITSPERPEL | DM_PELSWIDTH |
                      DM_PELSHEIGHT | DM_DISPLAYFREQUENCY |
                      DM_DISPLAYFLAGS;
  
  TSXFree(vidmodes);
  TRACE("returning: dmDisplayFlags = %ld dmDisplayFrequency = %ld, dmBitsPerPel = %ld,\n\tdmPelsWidth = %ld, dmPelsHeight = %ld\n",
                  devmode->dmDisplayFlags, devmode->dmDisplayFrequency, devmode->dmBitsPerPel,
                  devmode->dmPelsWidth, devmode->dmPelsHeight);
  return TRUE;
}

BOOL X11DRV_ChangeDisplayMode(LPDEVMODEA devmode)
{

  int i, do_enable = FALSE;
  int best = -1;
  static int originalmodeNum;
  HWND hwnd;
  BOOL got_mode = devmode && devmode->dmFields;

  if (got_mode) {
    TRACE("switching to %ldx%ldx%ld (%ld Hz)\n", devmode->dmPelsWidth, devmode->dmPelsHeight, devmode->dmBitsPerPel, devmode->dmDisplayFrequency);
    TRACE("fields %08lx\n", devmode->dmFields);
    TRACE("flags %08lx\n", devmode->dmDisplayFlags);
  } else {
    TRACE("switching to default mode\n");
  }

  if (got_mode && !app_resolution_change) {
    app_resolution_change = TRUE;
    do_enable = TRUE;
  } else
  if (!got_mode && app_resolution_change) {
    X11DRV_DisableGrabs();
    app_resolution_change = FALSE;
  }

/* if there are no modes available (eg the window is in a desktop), we return false
 * because we dont want ChangeDisplaySettings to move any windows around
 * since tehres no reason to.
 * But: ChangeDisplaySettings will always return SUCCESSFUL, because we want
 * the application to think that it was SUCCESSFUL.
 */
  if (!xf86vm_mode_count)
  {
    if (do_enable) X11DRV_EnableGrabs();
    return FALSE;
  }
  if (!got_mode && xf86vm_in_abnormalmode)
  {
    TRACE (" Changing back to default mode %i from %i\n", originalmodeNum, xf86vm_in_abnormalmode);
    xf86vm_in_abnormalmode = 0;
    X11DRV_XF86VM_SetCurrentMode(originalmodeNum);
    screen_width = default_screen_width;
    screen_height = default_screen_height;
    return FALSE; /* again why bother moving windows around?. This makes sure we dont */
  } /* if its null and we have never changed modes before, what they asking this for */
  else if (!got_mode)
    return FALSE;
  for (i = 0; i < xf86vm_mode_count; i++)
  {
    TRACE (" I found a mode: xf86vm_modes[%i] width = %ld\n", i, xf86vm_modes[i].dwWidth);
    /* we want to choose the smallest mode available that is still large enough for us
     * I'm not doing BPP or refresh rate matching, because its unneccesary. Just res is fine. */
    if (xf86vm_modes[i].dwWidth >= devmode->dmPelsWidth &&
        xf86vm_modes[i].dwHeight >= devmode->dmPelsHeight)
    {
      if (best == -1) best = i;
      else
      {
        if (xf86vm_modes[i].dwWidth < xf86vm_modes[best].dwWidth ||
            xf86vm_modes[i].dwHeight < xf86vm_modes[best].dwHeight)
	  best = i;
      }
    }
  }
  if (best == -1)
  {
    TRACE (" All modes available smaller than requested (%ldx%ld). Using current mode.\n", devmode->dmPelsWidth, devmode->dmPelsHeight);
    best = X11DRV_XF86VM_GetCurrentMode();
  }
  if (!xf86vm_in_abnormalmode)
  {
    originalmodeNum = X11DRV_XF86VM_GetCurrentMode();
  }
  xf86vm_in_abnormalmode = best;
  TRACE (" using mode: %d (%ldx%ld)\n", best, xf86vm_modes[best].dwWidth, xf86vm_modes[best].dwHeight);
  X11DRV_XF86VM_SetCurrentMode(best);
  
  /* Some Q3-engine games have an active console that we need to recenter before enabling the grab */
  hwnd = GetForegroundWindow();
  if (hwnd) {
    RECT rect;
    INT w, h, x, y;
    GetWindowRect(hwnd, &rect);
    w = rect.right - rect.left;
    h = rect.bottom - rect.top;
    if (w > xf86vm_modes[best].dwWidth) x = 0;
    else x = (xf86vm_modes[best].dwWidth - w) / 2;
    if (h > xf86vm_modes[best].dwHeight) y = 0;
    else y = (xf86vm_modes[best].dwHeight - h) / 2;
    MoveWindow(hwnd, x, y, w, h, TRUE);
  }
  
  if (do_enable) X11DRV_EnableGrabs();

  return TRUE;
}

void X11DRV_XF86VM_SetExclusiveMode(int lock)
{
  if (!xf86vm_modes) return; /* no XVidMode */

  TSXF86VidModeLockModeSwitch(gdi_display, DefaultScreen(gdi_display), lock);
}

/* actual DirectDraw HAL stuff */

static DWORD PASCAL X11DRV_XF86VM_SetMode(LPDDHAL_SETMODEDATA data)
{
  TRACE(" xvidmode switch from ddraw\n");
  X11DRV_XF86VM_SetCurrentMode(data->dwModeIndex);
  X11DRV_DDHAL_SwitchMode(data->dwModeIndex, NULL, NULL);
  data->ddRVal = DD_OK;
  return DDHAL_DRIVER_HANDLED;
}

int X11DRV_XF86VM_CreateDriver(LPDDHALINFO info)
{
  if (!xf86vm_mode_count) return 0; /* no XVidMode */

  info->dwNumModes = xf86vm_mode_count;
  info->lpModeInfo = xf86vm_modes;
  X11DRV_DDHAL_SwitchMode(X11DRV_XF86VM_GetCurrentMode(), NULL, NULL);
  info->lpDDCallbacks->SetMode = X11DRV_XF86VM_SetMode;
  return TRUE;
}


/***** GAMMA CONTROL *****/
/* (only available in XF86VidMode 2.x) */

#ifdef X_XF86VidModeSetGamma

static void GenerateRampFromGamma(WORD ramp[256], float gamma)
{
  float r_gamma = 1/gamma;
  unsigned i;
  TRACE("gamma is %f\n", r_gamma);
  for (i=0; i<256; i++)
    ramp[i] = pow(i/255.0, r_gamma) * 65535.0;
}

static BOOL ComputeGammaFromRamp(WORD ramp[256], float *gamma)
{
  float r_x, r_y, r_lx, r_ly, r_d, r_v, r_e, g_avg, g_min, g_max;
  unsigned i, f, l, g_n, c;
  f = ramp[0];
  l = ramp[255];
  if (f >= l) {
    ERR("inverted or flat gamma ramp (%d->%d), rejected\n", f, l);
    return FALSE;
  }
  r_d = l - f;
  g_min = g_max = g_avg = 0.0;
  /* check gamma ramp entries to estimate the gamma */
  TRACE("analyzing gamma ramp (%d->%d)\n", f, l);
  for (i=1, g_n=0; i<255; i++) {
    if (ramp[i] < f || ramp[i] > l) {
      ERR("strange gamma ramp ([%d]=%d for %d->%d), rejected\n", i, ramp[i], f, l);
      return FALSE;
    }
    c = ramp[i] - f;
    if (!c) continue; /* avoid log(0) */

    /* normalize entry values into 0..1 range */
    r_x = i/255.0; r_y = c / r_d;
    /* compute logarithms of values */
    r_lx = log(r_x); r_ly = log(r_y);
    /* compute gamma for this entry */
    r_v = r_ly / r_lx;
    /* compute differential (error estimate) for this entry */
    /* some games use table-based logarithms that magnifies the error by 128 */
    r_e = -r_lx * 128 / (c * r_lx * r_lx);

    /* compute min & max while compensating for estimated error */
    if (!g_n || g_min > (r_v + r_e)) g_min = r_v + r_e;
    if (!g_n || g_max < (r_v - r_e)) g_max = r_v - r_e;

    /* add to average */
    g_avg += r_v;
    g_n++;
    /* TRACE("[%d]=%d, gamma=%f, error=%f\n", i, ramp[i], r_v, r_e); */
  }
  if (!g_n) {
    ERR("no gamma data, shouldn't happen\n");
    return FALSE;
  }
  g_avg /= g_n;
  TRACE("low bias is %d, high is %d, gamma is %5.3f\n", f, 65535-l, g_avg);
  /* the bias could be because the app wanted something like a "red shift"
   * like when you're hit in Quake, but XVidMode doesn't support it,
   * so we have to reject a significant bias */
  if (f && f > (pow(1/255.0, g_avg) * 65536.0)) {
    ERR("low-biased gamma ramp (%d), rejected\n", f);
    return FALSE;
  }
  /* check that the gamma is reasonably uniform across the ramp */
  if (g_max - g_min > 0.1) {
    ERR("ramp not uniform (max=%f, min=%f, avg=%f), rejected\n", g_max, g_min, g_avg);
    return FALSE;
  }
  /* ok, now we're pretty sure we can set the desired gamma ramp,
   * so go for it */
  *gamma = 1/g_avg;
  return TRUE;
}

#endif /* X_XF86VidModeSetGamma */

/* Hmm... should gamma control be available in desktop mode or not?
 * I'll assume that it should */

BOOL X11DRV_XF86VM_GetGammaRamp(LPDDGAMMARAMP ramp)
{
#ifdef X_XF86VidModeSetGamma
  XF86VidModeGamma gamma;
  Bool ret;

  if (xf86vm_major < 2) return FALSE; /* no gamma control */
#ifdef X_XF86VidModeSetGammaRamp
  else if (xf86vm_use_gammaramp)
  {
      Bool ret;
      wine_tsx11_lock();
      ret = XF86VidModeGetGammaRamp(gdi_display, DefaultScreen(gdi_display), 256,
				    ramp->red, ramp->green, ramp->blue);
      wine_tsx11_unlock();
      return ret;
  }
#endif
  else
  {
      wine_tsx11_lock();
      ret = XF86VidModeGetGamma(gdi_display, DefaultScreen(gdi_display), &gamma);
      wine_tsx11_unlock();
      if (ret) {
	  GenerateRampFromGamma(ramp->red,   gamma.red);
	  GenerateRampFromGamma(ramp->green, gamma.green);
	  GenerateRampFromGamma(ramp->blue,  gamma.blue);
	  return TRUE;
      }
  }
#endif /* X_XF86VidModeSetGamma */
  return FALSE;
}

BOOL X11DRV_XF86VM_SetGammaRamp(LPDDGAMMARAMP ramp)
{
#ifdef X_XF86VidModeSetGamma
  XF86VidModeGamma gamma;

  if (xf86vm_major < 2) return FALSE; /* no gamma control */
#ifdef X_XF86VidModeSetGammaRamp
  else if (xf86vm_use_gammaramp)
  {
      Bool ret;
      wine_tsx11_lock();
      ret = XF86VidModeSetGammaRamp(gdi_display, DefaultScreen(gdi_display), 256,
				    ramp->red, ramp->green, ramp->blue);
      wine_tsx11_unlock();
      return ret;
  }
#endif
  else
  {
      if (ComputeGammaFromRamp(ramp->red,   &gamma.red) &&
	  ComputeGammaFromRamp(ramp->green, &gamma.green) &&
	  ComputeGammaFromRamp(ramp->blue,  &gamma.blue)) {
	  Bool ret;
	  wine_tsx11_lock();
	  ret = XF86VidModeSetGamma(gdi_display, DefaultScreen(gdi_display), &gamma);
	  wine_tsx11_unlock();
	  return ret;
      }
  }
#endif /* X_XF86VidModeSetGamma */
  return FALSE;
}

#endif /* HAVE_LIBXXF86VM */

/***********************************************************************
 *		GetDeviceGammaRamp (X11DRV.@)
 *
 * FIXME: should move to somewhere appropriate, but probably not before
 * the stuff in graphics/x11drv/ has been moved to dlls/x11drv, so that
 * they can include xvidmode.h directly
 */
BOOL X11DRV_GetDeviceGammaRamp(DC *dc, LPVOID ramp)
{
#ifdef HAVE_LIBXXF86VM
  return X11DRV_XF86VM_GetGammaRamp(ramp);
#else
  return FALSE;
#endif
}

/***********************************************************************
 *		SetDeviceGammaRamp (X11DRV.@)
 *
 * FIXME: should move to somewhere appropriate, but probably not before
 * the stuff in graphics/x11drv/ has been moved to dlls/x11drv, so that
 * they can include xvidmode.h directly
 */
BOOL X11DRV_SetDeviceGammaRamp(DC *dc, LPVOID ramp)
{
#ifdef HAVE_LIBXXF86VM
  return X11DRV_XF86VM_SetGammaRamp(ramp);
#else
  return FALSE;
#endif
}
